from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division
from concept_formation.cobweb3 import Cobweb3Tree
from concept_formation.cobweb3 import Cobweb3Node
from concept_formation.structure_mapper import structure_map

class TrestleTree(Cobweb3Tree):

    def __init__(self, alpha=0.001, scaling=True):
        self.root = Cobweb3Node()
        self.root.root = self.root
        self.root.alpha = alpha
        self.root.scaling = scaling

    def ifit(self, instance):
        """
        A modification of ifit to call Trestle instead.
        """
        return self.trestle(instance)

    def trestle_categorize(self, instance):
        """
        The Trestle categorize function, this Trestle categorizes all the
        sub-components before categorizing itself.
        """
        temp_instance = structure_map(self.root, instance)
        return self.cobweb_categorize(temp_instance)

    def categorize(self, instance):
        """
        A categorize function that can be used polymorphicaly without 
        having to worry about the type of the underlying object.

        In Trestle's case this calls trestle_categorize()
        """
        return self.trestle_categorize(instance)

    def trestle(self, instance):
        """
        Recursively calls Trestle on all of the components in a depth-first
        traversal. Once all of the components have been classified then then it
        classifies the current node.
        """
        temp_instance = structure_map(self.root, instance)
        return self.cobweb(temp_instance)



from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division
from concept_formation.cobweb3 import Cobweb3Node
from concept_formation.trestle import TrestleTree
from concept_formation.structure_mapper import flattenJSON
from concept_formation.structure_mapper import structure_map

class DummyTree(TrestleTree):

    def __init__(self):
        self.root = Cobweb3Node()
        self.root.root = self.root
        self.root.alpha = 0
        self.root.scaling = False

    def ifit(self, instance, do_mapping=False):
        """
        Just maintain a set of counts at the root and use these for prediction.

        The structure_map parameter determines whether or not to do structure
        mapping. This is disabled by default to get a really naive model.
        """
        if do_mapping:
            temp_instance = structure_map(self.root, instance)
        else:
            temp_instance = flattenJSON(instance)
        self.root.increment_counts(temp_instance)
        return self.root

    def categorize(self, instance):
        return self.root





from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division
from random import shuffle
from random import random
import json

from concept_formation.utils import weighted_choice

class CobwebTree:

    def __init__(self, alpha=0.001):
        """
        The tree constructor.

        The alpha parameter is the parameter used for laplacian smoothing. The
        higher the value, the higher the prior that all attributes/values are
        equally likely. By default a minor smoothing is used: 0.001.
        """
        self.root = CobwebNode()
        self.root.root = self.root
        self.root.alpha = alpha
        self.root.scaling = False

    def __str__(self):
        return str(self.root)

    def ifit(self, instance):
        """
        Given an instance incrementally update the categorization tree.
        """
        return self.cobweb(instance)

    def fit(self, instances, iterations=1, randomize_order=True):
        """
        This is a batch ifit function that takes a collection of instances
        and categorizes all of them. This function does not return anything.

        instances -- a collection of instances
        iterations -- the number of iterations to perform
        shuffle -- whether or not to shuffle the list between iterations.

        Note that the first iteration is not shuffled.
        """
        instances = [i for i in instances]

        for x in range(iterations):
            if randomize_order:
                shuffle(instances)

            for i in instances:
                self.ifit(i)

    def cobweb(self, instance):
        """
        Incrementally integrates an instance into the categorization tree.
        This function operates iteratively to integrate this instance and uses
        category utility as the heuristic to make decisions.
        """
        current = self.root

        while current:
            if (not current.children and current.cu_for_fringe_split(instance)
                <= 0.0):
                current.increment_counts(instance)
                return current 

            elif not current.children:
                new = current.__class__(current)
                current.parent = new
                new.children.append(current)

                if new.parent:
                    new.parent.children.remove(current)
                    new.parent.children.append(new)
                else:
                    new.alpha = self.root.alpha
                    new.scaling = self.root.scaling
                    self.root = new
                    self.root.root = new

                new.increment_counts(instance)
                return new.create_new_child(instance)
            else:
                best1, best2 = current.two_best_children(instance)
                action_cu, best_action = current.get_best_operation(instance,
                                                                    best1,
                                                                    best2)

                if best1:
                    best1_cu, best1 = best1
                if best2:
                    best2_cu, best2 = best2

                if best_action == 'best':
                    current.increment_counts(instance)
                    current = best1
                elif best_action == 'new':
                    current.increment_counts(instance)
                    return current.create_new_child(instance)
                elif best_action == 'merge':
                    current.increment_counts(instance)
                    new_child = current.merge(best1, best2)
                    current = new_child
                elif best_action == 'split':
                    current.split(best1)
                else:
                    raise Exception("Should never get here.")

    def cobweb_categorize(self, instance):
        """
        Sorts an instance in the categorization tree defined at the current
        node without modifying the counts of the tree.

        Uses the new and best operations; when new is the best operation it
        returns the current node otherwise it iterates on the best node. 
        """
        current = self.root
        while current:
            if not current.children:
                return current

            best1, best2 = current.two_best_children(instance)
            action_cu, best_action = current.get_best_operation(instance,
                                                                 best1, best2,
                                                                 ["best",
                                                                  "new"]) 
            if best1:
                best1_cu, best1 = best1
            else:
                return current

            if best_action == "new":
                return current
            elif best_action == "best":
                current = best1

    def categorize(self, instance):
        """
        A categorize function that can be used polymorphicaly without 
        having to worry about the type of the underlying object.

        In Cobweb's case this calls cobweb_categorize()
        """
        return self.cobweb_categorize(instance)
    
    def train_from_json(self, filename, length=None):
        """
        Build the concept tree from a set of examples in a provided json file.
        """
        json_data = open(filename, "r")
        instances = json.load(json_data)
        if length:
            shuffle(instances)
            instances = instances[:length]
        self.fit(instances)
        json_data.close()

class CobwebNode:

    counter = 0

    def __init__(self, otherNode=None):
        """
        The constructor creates a cobweb node with default values. It can also
        be used as a copy constructor to "deepcopy" a node.
        """
        self.concept_id = self.gensym()
        self.count = 0
        self.av_counts = {}
        self.children = []
        self.parent = None
        self.root = None

        if otherNode:
            self.update_counts_from_node(otherNode)
            self.parent = otherNode.parent
            self.root = otherNode.root

            for child in otherNode.children:
                self.children.append(self.__class__(child))

    def shallow_copy(self):
        """
        Creates a shallow copy of the current node (and not its children)
        """
        temp = self.__class__()
        temp.root = self.root
        temp.update_counts_from_node(self)
        return temp

    def increment_counts(self, instance):
        """
        Increment the counts at the current node according to the specified
        instance.

        input:
            instance: {a1: v1, a2: v2, ...} - a hashtable of attr and values. 
        """
        self.count += 1 
        for attr in instance:
            self.av_counts[attr] = self.av_counts.setdefault(attr,{})
            self.av_counts[attr][instance[attr]] = (self.av_counts[attr].get(
                instance[attr], 0) + 1)
    
    def update_counts_from_node(self, node):
        """
        Increments the counts of the current node by the amount in the specified
        node.
        """
        self.count += node.count
        for attr in node.av_counts:
            for val in node.av_counts[attr]:
                self.av_counts[attr] = self.av_counts.setdefault(attr,{})
                self.av_counts[attr][val] = (self.av_counts[attr].get(val,0) +
                                     node.av_counts[attr][val])

    def expected_correct_guesses(self):
        """
        Returns the number of correct guesses that are expected from the given
        concept. This is the sum of the probability of each attribute value
        squared. 
        """
        correct_guesses = 0.0

        for attr in self.root.av_counts:
            if attr[0] == "_":
                continue
            val_count = 0

            # the +1 is for the "missing" value
            n_values = len(self.root.av_counts[attr]) + 1

            for val in self.root.av_counts[attr]:
                if attr not in self.av_counts or val not in self.av_counts[attr]:
                    prob = 0
                    if self.root.alpha > 0:
                        prob = self.root.alpha / (self.root.alpha * n_values)
                else:
                    val_count += self.av_counts[attr][val]
                    prob = ((self.av_counts[attr][val] + self.root.alpha) / (1.0 * self.count
                                                              + self.root.alpha * n_values))
                correct_guesses += (prob * prob)

            #Factors in the probability mass of missing values
            prob = ((self.count - val_count + self.root.alpha) / (1.0*self.count +
                                                             self.root.alpha * n_values))
            correct_guesses += (prob * prob)

        return correct_guesses

    def category_utility(self):
        """
        Returns the category utility of a particular division of a concept into
        its children. This is used as the heuristic to guide the concept
        formation.
        """
        if len(self.children) == 0:
            return 0.0

        child_correct_guesses = 0.0

        for child in self.children:
            p_of_child = child.count / (1.0 * self.count)
            child_correct_guesses += p_of_child * child.expected_correct_guesses()

        return ((child_correct_guesses - self.expected_correct_guesses()) /
                (1.0 * len(self.children)))

    def get_best_operation(self, instance, best1, best2, 
                            possible_ops=["best", "new", "merge", "split"]):
        """
        Given a set of possible operations, find the operator that produces the
        highest category utility, and then returns the cu and the action name
        for the best operation. In the case of ties, an operator is randomly
        chosen.
        """
        if not best1:
            raise ValueError("Need at least one best child.")

        if best1:
            best1_cu, best1 = best1
        if best2:
            best2_cu, best2 = best2
        operations = []

        if "best" in possible_ops:
            operations.append((best1_cu, random(), "best"))
        if "new" in possible_ops: 
            operations.append((self.cu_for_new_child(instance), random(), 'new'))
        if "merge" in possible_ops and len(self.children) > 2 and best2:
            operations.append((self.cu_for_merge(best1, best2, instance),
                               random(),'merge'))
        if "split" in possible_ops and len(best1.children) > 0:
            operations.append((self.cu_for_split(best1), random(), 'split'))

        operations.sort(reverse=True)
        #print(operations)
        best_op = (operations[0][0], operations[0][2])
        #print(best_op)
        return best_op

    def two_best_children(self, instance):
        """
        Returns the two best children to incorporate the instance
        into in terms of category utility. When selecting the best children it
        sorts first by CU, then by Size, then randomly. 

        input:
            instance: {a1: v1, a2: v2,...} - a hashtable of attr. and values. 
        output:
            (0.2,2),(0.1,3) - the category utility and indices for the two best
            children (the second tuple will be None if there is only 1 child).
        """
        if len(self.children) == 0:
            raise Exception("No children!")

        children_cu = [(self.cu_for_insert(child, instance), child.count,
                        random(), child) for child in self.children]
        children_cu.sort(reverse=True)

        if len(children_cu) == 0:
            return None, None
        if len(children_cu) == 1:
            return (children_cu[0][0], children_cu[0][3]), None 

        return ((children_cu[0][0], children_cu[0][3]), (children_cu[1][0],
                                                         children_cu[1][3]))

    def cu_for_insert(self, child, instance):
        """
        Computer the category utility of adding the instance to the specified
        child.
        """
        temp = self.shallow_copy()
        temp.increment_counts(instance)

        for c in self.children:
            temp_child = c.shallow_copy()
            temp.children.append(temp_child)
            if c == child:
                temp_child.increment_counts(instance)
        return temp.category_utility()

    def create_new_child(self, instance):
        """
        Creates a new child (to the current node) with the counts initialized by
        the given instance. 
        """
        new_child = self.__class__()
        new_child.parent = self
        new_child.root = self.root
        new_child.increment_counts(instance)
        self.children.append(new_child)
        return new_child

    def create_child_with_current_counts(self):
        """
        Creates a new child (to the current node) with the counts initialized by
        the current node's counts.
        """
        if self.count > 0:
            new = self.__class__(self)
            new.parent = self
            new.root = self.root
            self.children.append(new)
            return new

    def cu_for_new_child(self, instance):
        """
        Returns the category utility for creating a new child using the
        particular instance.
        """
        temp = self.shallow_copy()
        for c in self.children:
            temp.children.append(c.shallow_copy())

        #temp = self.shallow_copy()
        
        temp.increment_counts(instance)
        temp.create_new_child(instance)
        return temp.category_utility()

    def merge(self, best1, best2):
        """
        Merge the two specified nodes.

        input:
            best1: the best child
            best2: the second best child
        output:
            The new child formed from the merge
        """
        new_child = self.__class__()
        new_child.parent = self
        new_child.root = self.root
        new_child.update_counts_from_node(best1)
        new_child.update_counts_from_node(best2)
        best1.parent = new_child
        best1.root = new_child.root
        best2.parent = new_child
        best2.root = new_child.root
        new_child.children.append(best1)
        new_child.children.append(best2)
        self.children.remove(best1)
        self.children.remove(best2)
        self.children.append(new_child)

        return new_child

    def cu_for_merge(self, best1, best2, instance):
        """
        Returns the category utility for merging the two best children.

        input:
            best1: the best child in the children array.
            best2: the second best child in the children array.
        output:
            0.02 - the category utility for the merge of best1 and best2.
        """
        temp = self.shallow_copy()
        temp.increment_counts(instance)

        new_child = self.__class__()
        new_child.root = self.root
        new_child.update_counts_from_node(best1)
        new_child.update_counts_from_node(best2)
        new_child.increment_counts(instance)
        temp.children.append(new_child)

        for c in self.children:
            if c == best1 or c == best2:
                continue
            temp_child = c.shallow_copy()
            temp.children.append(temp_child)

        return temp.category_utility()

    def split(self, best):
        """
        Split the best node and promote its children
        """
        self.children.remove(best)
        for child in best.children:
            child.parent = self
            child.root = self.root
            self.children.append(child)

    def cu_for_fringe_split(self, instance):
        """
        Determine the category utility of performing a fringe split (i.e.,
        adding a leaf to a leaf). It turns out that this is useful for
        identifying unnecessary fringe splits, when the two leaves are
        essentially identical. It can be used to keep the tree from growing and
        to increase the tree's predictive accuracy.
        """
        temp = self.shallow_copy()
        
        temp.create_child_with_current_counts()
        temp.increment_counts(instance)
        temp.create_new_child(instance)

        return temp.category_utility()

    def cu_for_split(self, best):
        """
        Return the category utility for splitting the best child.
        
        input:
            best1: a child in the children array.
        output:
            0.03 - the category utility for the split of best1.
        """
        temp = self.shallow_copy()

        for c in self.children + best.children:
            if c == best:
                continue
            temp_child = c.shallow_copy()
            temp.children.append(temp_child)

        return temp.category_utility()

    def __hash__(self):
        """
        The basic hash function. This hashes the concept name, which is
        generated to be unique across concepts.
        """
        return hash("CobwebNode" + str(self.concept_id))

    def gensym(self):
        """
        Generates a unique id and increments the class counter. This is used to
        create a unique name for every concept. 
        """
        self.__class__.counter += 1
        return str(self.__class__.counter)

    def __str__(self):
        """
        Call pretty print
        """
        return self.pretty_print()

    def pretty_print(self, depth=0):
        """
        Prints the categorization tree.
        """
        ret = str(('\t' * depth) + "|-" + str(self.av_counts) + ":" +
                  str(self.count) + '\n')
        
        for c in self.children:
            ret += c.pretty_print(depth+1)

        return ret

    def depth(self):
        """
        Returns the depth of the current node in the tree.
        """
        if self.parent:
            return 1 + self.parent.depth()
        return 0

    def is_parent(self, other_concept):
        """
        Returns True if self is a parent of other concept.
        """
        temp = other_concept
        while temp != None:
            if temp == self:
                return True
            try:
                temp = temp.parent
            except:
                print(temp)
                assert False
        return False

    def num_concepts(self):
        """
        Return the number of concepts contained in the tree defined by the
        current node. 
        """
        children_count = 0
        for c in self.children:
           children_count += c.num_concepts() 
        return 1 + children_count 

    def output_json(self):
        """
        Outputs the categorization tree in JSON form so that it can be
        displayed, I usually visualize it with d3js in a web browser.
        """
        output = {}
        output['name'] = "Concept" + self.concept_id
        output['size'] = self.count
        output['children'] = []

        temp = {}
        for attr in self.av_counts:
            for value in self.av_counts[attr]:
                temp[attr + " = " + str(value)] = self.av_counts[attr][value]

        for child in self.children:
            output['children'].append(child.output_json())

        output['counts'] = temp

        return output

    def get_weighted_values(self, attr):
        n_values = len(self.root.av_counts[attr]) + 1

        choices = []
        val_count = 0
        for val in self.root.av_counts[attr]:
            if attr not in self.av_counts or val not in self.av_counts[attr]:
                count = 0
            else:
                count = self.av_counts[attr][val]

            choices.append((val, (count + self.root.alpha)
                            / (1.0 * self.count + self.root.alpha * n_values)))
            if attr in self.av_counts and val in self.av_counts[attr]:
                val_count += self.av_counts[attr][val]

        choices.append((None, ((self.count - val_count + self.root.alpha) /
                               (1.0 * self.count + self.root.alpha *
                                n_values))))
        return choices

    def predict(self, attr):
        """
        Predicts the value of an attribute, by returning the most likely value.
        This takes into account the laplacian smoothing. 
        """
        if attr not in self.root.av_counts:
            return None

        choices = self.get_weighted_values(attr)
        choices.sort(key=lambda x: -x[1])
        return choices[0][0]

    def sample(self, attr):
        """
        Predicts the value of an attribute, by returning the most likely value.
        This takes into account the laplacian smoothing. 
        """
        if attr not in self.root.av_counts:
            return None

        choices = self.get_weighted_values(attr)

        return weighted_choice(choices)

    def get_probability(self, attr, val):
        """
        Gets the probability of a particular attribute value at the given
        concept. This takes into account the possibilities that an attribute
        can take any of the values available at the root, or be missing.
        Laplace smoothing is used to place a prior over these possibilites.
        Alpha determines the strength of this prior.
        """
        if attr not in self.root.av_counts:
            return 0.0

        if val is not None and val not in self.root.av_counts[attr]:
            return 0.0

        n_values = len(self.root.av_counts[attr]) + 1

        count = 0
        if attr in self.av_counts and val in self.av_counts[attr]:
            count = self.av_counts[attr][val]

        return ((count + self.root.alpha) / 
                (1.0 * self.count + self.root.alpha * n_values))

    def get_probability_missing(self, attr):
        """
        Gets the probability of a particular attribute value at the given
        concept. This takes into account the possibilities that an attribute
        can take any of the values available at the root, or be missing.
        Laplace smoothing is used to place a prior over these possibilites.
        Alpha determines the strength of this prior.
        """
        # the +1 is for the "missing" value
        if attr in self.root.av_counts:
            n_values = len(self.root.av_counts[attr]) + 1
        else:
            n_values = 1

        val_count = 0
        if attr in self.av_counts:
            for val in self.av_counts[attr]:
                val_count += self.av_counts[attr][val]

        if (1.0 * self.count + self.root.alpha * n_values) == 0:
            return 0.0

        return ((self.count - val_count + self.root.alpha) / 
                (1.0 * self.count + self.root.alpha * n_values))




from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division

from concept_formation import search

gensym_counter = 0;

def gensym():
    """
    Generates unique names for naming renaming apart objects.
    """
    global gensym_counter
    gensym_counter += 1
    return 'o' + str(gensym_counter)

def standardizeApartNames(instance):
    """
    Given an raw input instance (i.e., relations are still lists, object is
    still structured), it renames all the components so they have unique names.

    >>> import pprint
    >>> pprint.pprint(standardizeApartNames({'a': {}, 'r1': ['is-good', 'a']}))
    {'o2': {}, 'r1': ['is-good', 'o2']}
    """
    new_instance = {}
    relations = []
    mapping = {}

    for attr in instance:
        if isinstance(instance[attr], list):
            relations.append((attr, instance[attr]))
        elif isinstance(instance[attr], dict):
            mapping[attr] = gensym()
            new_instance[mapping[attr]] = standardizeApartNames(instance[attr])
        else:
            new_instance[attr] = instance[attr]

    for name, r in relations:
        new_relation = []
        for i,v in enumerate(r):
            if i == 0:
                new_relation.append(v)
            else:
                new_relation.append(mapping[v])
        new_instance[name] = new_relation

    return new_instance

def getComponentNames(instance):
    """
    Given a flattened representation of an instance or concept.av_counts
    return a list of all of the component names.

    >>> sorted(getComponentNames({'c1.a': 0, 'c2.a': 0, '_c3._a': 0}))
    ['c1', 'c2', 'c3']
    """
    names = set()
    for attr in instance:
        if isinstance(attr, tuple):
            continue

        for name in attr.split(".")[:-1]:
            if name[0] == "_":
                names.add(name[1:])
            else:
                names.add(name)

    return list(names)

def renameComponent(attr, mapping):
    """
    Takes a component attribute (e.g., o1.o2) and renames the 
    components given a mapping.

    >>> renameComponent("c1.c2.a", {'c1': 'o1', 'c2': 'o2'})
    'o1.o2.a'

    >>> renameComponent("_c1._c2._a", {'c1': 'o1', 'c2': 'o2'})
    '_o1._o2._a'
    """
    new_attr = []
    att_split = attr.split('.')
    for name in att_split[:-1]:
        if name[0] == "_":
            new_attr.append(mapping[name[1:]])
        else:
            new_attr.append(mapping[name])

    if att_split[-1][0] == "_":
        new_attr.append(att_split[-1][1:])
    else:
        new_attr.append(att_split[-1])

    if attr[0] == "_":
        return "_" + "._".join(new_attr)
    else:
        return ".".join(new_attr)

def renameRelation(attr, mapping):
    """
    Takes a relational attribute (e.g., (before o1 o2)) and renames
    the components based on mapping.

    >>> renameRelation(('before', 'c1', 'c2'),  {'c1': 'o1', 'c2': 'o2'})
    ('before', 'o1', 'o2')
    """
    temp = []
    for idx, val in enumerate(attr):
        if idx == 0:
            temp.append(val)
        else:
            new_attr = []
            for name in val.split("."):
                new_attr.append(mapping[name])
            temp.append(".".join(new_attr))
    return tuple(temp)

def renameFlat(instance, mapping):
    """
    Given a flattened instance and a mapping (type = dict) rename the
    components and relations and return the renamed instance. 

    >>> import pprint
    >>> pprint.pprint(renameFlat({'c1.a': 1, ('good', 'c1'): True}, {'c1': 'o1'}))
    {'o1.a': 1, ('good', 'o1'): True}
    """
    for attr in instance:
        if isinstance(attr, tuple):
            continue
        for name in attr.split('.')[:-1]:
            if name[0] == "_":
                name = name[1:]
            if name not in mapping:
                mapping[name] = name

    temp_instance = {}

    for attr in instance:
        if isinstance(attr, tuple):
            temp_instance[renameRelation(attr, mapping)] = instance[attr]
        elif "." in attr:
            temp_instance[renameComponent(attr, mapping)] = instance[attr]
        else:
            temp_instance[attr] = instance[attr]

    return temp_instance

def flattenJSON(instance):
    """
    Takes a raw hierarchical JSON instance, standardizes apart the component
    names, and flattens it. It represents hierarchy with periods between variable
    names. It also converts the relations into tuples with values.

    >>> import pprint
    >>> pprint.pprint(flattenJSON({'a': 1, 'c1': {'b': 1}}))
    {'a': 1, 'o1.b': 1}
    """
    instance = standardizeApartNames(instance)
    temp = {}
    for attr in instance:
        if isinstance(instance[attr], dict):
            subobject = flattenJSON(instance[attr])
            for so_attr in subobject:
                if isinstance(so_attr, tuple):
                    relation = []
                    for idx, val in enumerate(so_attr):
                        if idx == 0:
                            relation.append(val)
                        else:
                            relation.append(attr + "." + val)
                    temp[tuple(relation)] = True
                elif so_attr[0] == "_":
                    temp["_" + attr + "." + so_attr] = subobject[so_attr]
                else:
                    temp[attr + "." + so_attr] = subobject[so_attr]
        else:
            if isinstance(instance[attr], list):
                temp[tuple(instance[attr])] = True
            else:
                temp[attr] = instance[attr]
    return temp

def traverseStructure(path, instance):
    """
    Given an instance (hashmap) to the given subobject for the given path.
    Creates subobjects if they do not exist.

    >>> x = {}
    >>> traverseStructure(['c1', 'c2'], x)
    {}
    >>> x
    {'c1': {'c2': {}}}
    """
    curr = instance
    for obj in path:
        if obj not in curr:
            curr[obj] = {}
        curr = curr[obj]
    return curr

def structurizeJSON(instance):
    """
    Takes a flattened instance and adds the structure back in. This essentially
    "undoes" the flattening process. 

    >>> structurizeJSON({'c1.c2.a': 1})
    {'c1': {'c2': {'a': 1}}}
    """
    temp = {}
    for attr in instance:
        if isinstance(attr, tuple):
            relation = []
            path = []
            for i,v in enumerate(attr):
                if i == 0:
                    relation.append(v)
                else:
                    path = v.split('.')
                    relation.append(path[-1])
                    path = path[:-1]
            obj = traverseStructure(path, temp)
            obj[tuple(relation)] = True

        elif "." in attr:
            path = [p[1:] if p[0] == "_" else p for p in attr.split('.')]
            subatt = path[-1]
            path = path[:-1]
            curr = traverseStructure(path, temp)
            if attr[0] == "_":
                curr["_" + subatt] = instance[attr]
            else:
                curr[subatt] = instance[attr]

        else:
            temp[attr] = instance[attr]

    return temp

def bindFlatAttr(attr, mapping):
    """
    Renames the attribute given the mapping.

    >>> bindFlatAttr(('before', 'c1', 'c2'), {'c1': 'o1', 'c2':'o2'})
    ('before', 'o1', 'o2')

    If the mapping is incomplete then returns None (nothing) 
    >>> bindFlatAttr(('before', 'c1', 'c2'), {'c1': 'o1'}) is None
    True
    """
    if isinstance(attr, tuple):
        for i,v in enumerate(attr):
            if i == 0:
                continue
            for o in v.split('.'):
                if o not in mapping:
                    return None
        return renameRelation(attr, mapping)
    elif '.' in attr:
        path = attr.split('.')[:-1]
        for o in path:
            if o not in mapping:
                return None
        return renameComponent(attr, mapping)
    else:
        return attr

def containsComponent(component, attr):
    """
    Check if the given component is in the attribute.
    
    >>> containsComponent('c1', 'c2.c1.a')
    True

    >>> containsComponent('c3', ('before', 'c1', 'c2'))
    False
    """
    if isinstance(attr, tuple):
        for i,v in enumerate(attr):
            if i == 0:
                continue
            for o in v.split('.'):
                if o == component:
                    return True
    elif "." in attr:
        for o in attr.split('.')[:-1]:
            if o == component:
                return True

    return False

def flatMatch(concept, instance, optimal=False):
    """
    Given a concept and instance this function returns a mapping (dictionary)
    that can be used to rename the instance. The mapping returned maximizes
    similarity between the instance and the concept.
    """
    inames = frozenset(getComponentNames(instance))
    cnames = frozenset(getComponentNames(concept.av_counts))

    if(len(inames) == 0 or
       len(cnames) == 0):
        return {}
     
    initial = search.Node((frozenset(), inames, cnames), extra=(concept, instance))
    if optimal:
        solution = next(search.BestFGS(initial, flatMatchSuccessorFn, flatMatchGoalTestFn,
                                flatMatchHeuristicFn))
    else:
        solution = next(search.BeamGS(initial, flatMatchSuccessorFn, flatMatchGoalTestFn,
                           flatMatchHeuristicFn, initialBeamWidth=3))
    #print(solution.cost)

    if solution:
        mapping, unnamed, availableNames = solution.state
        return {a:v for a,v in mapping}
    else:
        return None

def flatMatchSuccessorFn(node):
    """
    Given a node (mapping, instance, concept), this function computes the
    successor nodes where an additional mapping has been added for each
    possible additional mapping. See the search library for more details.
    """
    mapping, inames, availableNames = node.state
    concept, instance = node.extra

    for n in inames:
        reward = 0
        m = {a:v for a,v in mapping}
        m[n] = n
        for attr in instance:
            if not containsComponent(n, attr):
                continue
            new_attr = bindFlatAttr(attr, m)
            if new_attr:
                reward += concept.attr_val_guess_gain(new_attr, instance[attr])

        yield search.Node((mapping.union(frozenset([(n, n)])), inames -
                    frozenset([n]), availableNames), node, n + ":" + n,
                   node.cost - reward, node.depth + 1, node.extra)

        for new in availableNames:
            reward = 0
            m = {a:v for a,v in mapping}
            m[n] = new
            for attr in instance:
                if not containsComponent(n, attr):
                    continue
                new_attr = bindFlatAttr(attr, m)
                if new_attr:
                    reward += concept.attr_val_guess_gain(new_attr,
                                                          instance[attr])
            yield search.Node((mapping.union(frozenset([(n, new)])), inames -
                                      frozenset([n]), availableNames -
                                      frozenset([new])), node, n + ":" + new,
                        node.cost - reward, node.depth + 1, node.extra)

def flatMatchHeuristicFn(node):
    """
    Considers all partial matches for each unbound attribute and assumes that
    you get the highest guess_gain match. This provides an over estimation of
    the possible reward (i.e., is admissible).
    """
    mapping, unnamed, availableNames = node.state
    concept, instance = node.extra

    h = 0
    m = {a:v for a,v in mapping}
    for attr in instance:
        new_attr = bindFlatAttr(attr, m)
        if not new_attr:
            best_attr_h = [concept.attr_val_guess_gain(cAttr, instance[attr]) for
                               cAttr in concept.av_counts if
                               isPartialMatch(attr, cAttr, m)]

            if len(best_attr_h) != 0:
                h -= max(best_attr_h)

    return h

def flatMatchGoalTestFn(node):
    """
    Returns True if every component in the original instance has been renamed
    in the given node.
    """
    mapping, unnamed, availableNames = node.state
    return len(unnamed) == 0

def isPartialMatch(iAttr, cAttr, mapping):
    """
    Returns True if the instance attribute (iAttr) partially matches the
    concept attribute (cAttr) given the mapping.
    """
    if type(iAttr) != type(cAttr):
        return False
    if isinstance(iAttr, tuple) and len(iAttr) != len(cAttr):
        return False

    if isinstance(iAttr, tuple):
        if iAttr[0] != cAttr[0]:
            return False
        for i,v in enumerate(iAttr):
            if i == 0:
                continue
            iSplit = v.split('.')
            cSplit = cAttr[i].split('.')
            if len(iSplit) != len(cSplit):
                return False
            for j,v2 in enumerate(iSplit):
                if v2 in mapping and mapping[v2] != cSplit[j]:
                    return False
    elif "." not in cAttr:
        return False
    else:
        iSplit = iAttr.split('.')
        cSplit = cAttr.split('.')
        if len(iSplit) != len(cSplit):
            return False
        if iSplit[-1] != cSplit[-1]:
            return False
        for i,v in enumerate(iSplit[:-1]):
            if v in mapping and mapping[v] != cSplit[i]:
                return False

    return True

def structure_map(concept, instance):
    """
    Flatten the instance, perform structure mapping, rename the instance
    based on this structure mapping, and return the renamed instance.
    """
    temp_instance = flattenJSON(instance)
    mapping = flatMatch(concept, temp_instance)
    temp_instance = renameFlat(temp_instance, mapping)
    return temp_instance



from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division
from random import uniform
from numbers import Number
from math import sqrt

from scipy.stats import sem
from scipy.stats import t
from scipy import linalg
import numpy as np

# A hashtable of values to use in the c4(n) function to apply corrections to
# estimates of std.
c4n_table = {2: 0.7978845608028654, 
      3:  0.886226925452758, 
      4:  0.9213177319235613, 
      5:  0.9399856029866254, 
      6:  0.9515328619481445, 
      7:  0.9593687886998328, 
      8:  0.9650304561473722, 
      9:  0.9693106997139539, 
      10: 0.9726592741215884, 
      11: 0.9753500771452293, 
      12: 0.9775593518547722, 
      13: 0.9794056043142177, 
      14: 0.9809714367555161, 
      15: 0.9823161771626504, 
      16: 0.9834835316158412, 
      17: 0.9845064054718315, 
      18: 0.985410043808079, 
      19: 0.9862141368601935, 
      20: 0.9869342675246552, 
      21: 0.9875829288261562, 
      22: 0.9881702533158311, 
      23: 0.988704545233999, 
      24: 0.9891926749585048, 
      25: 0.9896403755857028, 
      26: 0.9900524688409107, 
      27: 0.990433039209448, 
      28: 0.9907855696217323, 
      29: 0.9911130482419843}

def c4(n) :
    """
    Returns the correction factor to apply to unbias estimates of standard 
    deviation in low sample sizes. This implementation is based on a lookup 
    table for n in [2-29] and returns 1.0 for values >= 30.

    >>> c4(3)
    0.886226925452758
    """
    if n <= 1 :
        raise ValueError("Cannot apply correction for a sample size of 1.")
    else :
        return c4n_table[n] if n < 30 else 1.0

def mean(values):
    """
    Computes the mean of a list of values.

    >>> mean([600, 470, 170, 430, 300])
    394.0
    """
    if len(values) <= 0:
        raise ValueError("Length of list must be greater than 0.")

    return float(sum(values))/len(values)

def std(values):
    """
    Computes the standard deviation of a list of values.

    >>> std([600, 470, 170, 430, 300])
    147.32277488562318
    """
    if len(values) <= 0:
        raise ValueError("Length of list must be greater than 0.")

    meanValue = mean(values)
    variance =  float(sum([(v - meanValue) * (v - meanValue) for v in
                           values]))/len(values)
    return sqrt(variance)

def listsToRelations(instance, relationName="Ordered", appendAttr=True, attrs = None):
    """
    Takes a structured instance containing lists and returns the same instance
    with all list elements converted to a series of relations that describe the
    ordering of elements. This process will be applied recurrsively to any
    commponent attributes of the instance.

    Arguments: 

        relationName -- The name that should be used to describe the relation,
        by default "Orderd" is used. However, a new name can be provided if
        that conflicts with other relations in the data already.

        appendAttr -- if True appends the originzal list's attribute name to the
        relationName this is to prevent the matcher from over generalizing
        ordering when there are multiple lists in an object.

        attrs -- A list of specific attribute names to convert. If no list is
        provided then all list attributes will be converted

    >>> import pprint
    >>> pprint.pprint(listsToRelations({"list1":['a','b','c']}))
    {'list1-0': ['Orderedlist1', 'a', 'b'], 'list1-1': ['Orderedlist1', 'b', 'c']}

    >>> import pprint
    >>> pprint.pprint(listsToRelations({"list1":['a','b','c']}, appendAttr=False))
    {'list1-0': ['Ordered', 'a', 'b'], 'list1-1': ['Ordered', 'b', 'c']}
    """
    if attrs is None:
        attrs = [k for k in instance.keys()]

    if isinstance(attrs,str):
        attrs = [attrs]

    newInstance = {}
    for attr in instance:
        if isinstance(instance[attr], list) and attr in attrs:
            for i in range(len(instance[attr])-1):
                if appendAttr:
                    newInstance[str(attr)+"-"+str(i)] = [str(relationName+attr),
                        instance[attr][i],
                        instance[attr][i+1]]
                else:
                    newInstance[str(attr)+"-"+str(i)] = [str(relationName),
                        instance[attr][i],
                        instance[attr][i+1]]
        elif isinstance(instance[attr],dict):
            newInstance[attr] = listsToRelations(instance[attr],relationName,appendAttr,attrs)
        else:
            newInstance[attr] = instance[attr]
    return newInstance

def batchListsToRelations(instances, relationName="Ordered", appendAttr=True, attrs=None):
    """
    Takes a list of structured instances that contain lists and batch converts
    all the list elements to a relation format expected by TRESTLE.
    """
    return [listsToRelations(instance, relationName, appendAttr) for instance
            in instances]

def extractComponentsFromList(instance, prefix="exob", attrs=None):
    """

    Extracts object literals from within an list elements and extracts them into
    their own component attributes. Note that this function does not ad ordering
    relations so if that behavior is desired follow it up with listToRelation.

    Arguments:

        prefix -- Any objects found within list in the ojbect will be assigned
        new names of prefix + a sequential counter. This value will eventually
        be aliased by TRESTLE's structure mapper but it should not collide with
        any object names that naturally exist wihtin the instance object.

        attrs -- A list of attribute names to specifically target for object
        extraction. If this is not provided then the process will search within
        all of the instance's attribtues.

    >>> import pprint
    >>> pprint.pprint(extractComponentsFromList({"a":"n","l1":["test",{"p":"q","j":"k"},{"n":"m"}]}))
    {'a': 'n',
     'exob1': {'j': 'k', 'p': 'q'},
     'exob2': {'n': 'm'},
     'l1': ['test', 'exob1', 'exob2']}

    >>> import pprint
    >>> pprint.pprint(extractComponentsFromList({"a":"n","l1":["test",{"p":"q","j":"k"},{"n":"m"}]},prefix="newobject"))
    {'a': 'n',
     'l1': ['test', 'newobject1', 'newobject2'],
     'newobject1': {'j': 'k', 'p': 'q'},
     'newobject2': {'n': 'm'}}
 
    """
    counter = 0

    if attrs is None:
        attrs = [k for k in instance.keys()]

    if isinstance(attrs,str):
        attrs = [attrs]

    for a in attrs:
        if  isinstance(instance[a],list):
            for i in range(len(instance[a])):
                if isinstance(instance[a][i],dict):
                    counter += 1
                    newName = prefix + str(counter)
                    instance[newName] = extractComponentsFromList(instance[a][i])
                    instance[a][i] = newName
        if isinstance(instance[a],dict):
            instance[a] = extractComponentsFromList(instance[a])
    return instance

def batchExtractComponentsFromList(instances, prefix="exob", attrs=None):
    """

    Takes a list of instances and extracts components from lists within each of
    them.

    """
    return [extractComponentsFromList(instance, prefix, attrs) for instance
            in instances]

def numericToNominal(instance, attrs = None):     
    """         
    Takes a list of instances and converts any attributes that TRESTLE or
    COBWEB/3 would consider to be numeric attributes to nominal attributes in
    the case that should be treated as such. This is useful for when data
    natually contains values that would be numbers but you do not want to
    entertain that they have a distribution.    

    Arguments:

        attrs -- a list of attribute names to specifically target for
        conversion. If no list is provided it will default to all of the
        instance's attributes.

    >>> import pprint
    >>> pprint.pprint(numericToNominal({"x":12.5,"y":9,"z":"top"}))
    {'x': '12.5', 'y': '9', 'z': 'top'}

    >>> import pprint
    >>> pprint.pprint(numericToNominal({"x":12.5,"y":9,"z":"12.6"},attrs=["y","z"]))
    {'x': 12.5, 'y': '9', 'z': '12.6'}

    >>> import pprint
    >>> pprint.pprint(numericToNominal({"x":12.5,"y":9,"z":"top"},attrs="y"))
    {'x': 12.5, 'y': '9', 'z': 'top'}

    """
    if attrs is None:
        attrs = [k for k in instance.keys()]
    if isinstance(attrs,str):
        attrs = [attrs]
    for a in attrs:
        if isinstance(instance[a],Number):
            instance[a] = str(instance[a])
        if isinstance(instance[a],dict):
            instance[a] = numericToNominal(instance[a],attrs)
    return instance

def batchNumericToNominal(instances, attrs = None):
    """
    Takes a list of instances and batch converts any specified numeric
    attributes within instances to nominal attributes for times when this is the
    desired behavior
    """
    return [numericToNominal(instance,attrs) for instance in instances]

def santizeJSON(instance,ext_ob_name="extOb",relation_name="Ordered",appendAttr=True,attrs = None):
    """

    Takes a raw JSON object and applies some transformations to it that convert
    common structures into an equivalent format that TRESTLE would expect.
    TRESTLE should run given any arbitrarily structured JSON object but this
    process makes some conversions that preserve the intent of some JSON
    notation in a way that TRESTLE would expect. In particular this process will
    extract an object litterals that are in-line within a list, and break any
    lists into a series of tuples that describe their order.

    Attributes:

        ext_ob_name -- the prefix to use for objects extracted from in-line lists

        relation_name -- the name to use for ordering relations converted from lists

        appendAttr -- a flag for whether to append the original attribute name to a converted relation list

        attrs -- a list of particular attributes to perform conversion on.

    >>> import pprint
    >>> pprint.pprint(santizeJSON({"stack":[{"a":1,"b":2,"c":3},{"x":1,"y":2,"z":3},{"i":1,"j":2,"k":3}]}))
    {'extOb1': {'a': 1, 'b': 2, 'c': 3},
     'extOb2': {'x': 1, 'y': 2, 'z': 3},
     'extOb3': {'i': 1, 'j': 2, 'k': 3},
     'stack-0': ['Orderedstack', 'extOb1', 'extOb2'],
     'stack-1': ['Orderedstack', 'extOb2', 'extOb3']}

    """    

    return listsToRelations(extractComponentsFromList(instance,prefix=ext_ob_name,
        attrs=attrs),relationName=relation_name,appendAttr=appendAttr,attrs=attrs)

def santizeJSONList(instances,ext_ob_name="extOb",relation_name="Ordered",appendAttr=True,attrs = None):
    """

    Applies the santizeJSON function to a list of objects.

    """
    return [santizeJSON(instance,ext_ob_name,relation_name,appendAttr,attrs) for instance in instances]

def moving_average(a, n=3) :
    """
    A function for computing the moving average, so that we can smooth out the
    curves on a graph.
    """
    ret = np.cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n

def lowess(x, y, f=1./3., iter=3, confidence=0.95):
    """
    Code taken from: https://gist.github.com/agramfort/850437

    lowess(x, y, f=2./3., iter=3) -> yest

    Lowess smoother: Robust locally weighted regression.
    The lowess function fits a nonparametric regression curve to a scatterplot.
    The arrays x and y contain an equal number of elements; each pair
    (x[i], y[i]) defines a data point in the scatterplot. The function returns
    the estimated (smooth) values of y.

    The smoothing span is given by f. A larger value for f will result in a
    smoother curve. The number of robustifying iterations is given by iter. The
    function will run faster with a smaller number of iterations.
    """
    n = len(x)
    r = int(np.ceil(f*n))
    h = [np.sort(np.abs(x - x[i]))[r] for i in range(n)]
    w = np.clip(np.abs((x[:,None] - x[None,:]) / h), 0.0, 1.0)
    w = (1 - w**3)**3
    yest = np.zeros(n)
    delta = np.ones(n)
    for iteration in range(iter):
        for i in range(n):
            weights = delta * w[:,i]
            b = np.array([np.sum(weights*y), np.sum(weights*y*x)])
            A = np.array([[np.sum(weights), np.sum(weights*x)],
                   [np.sum(weights*x), np.sum(weights*x*x)]])
            beta = linalg.solve(A, b)
            yest[i] = beta[0] + beta[1]*x[i]
 
        residuals = y - yest
        s = np.median(np.abs(residuals))
        delta = np.clip(residuals / (6.0 * s), -1, 1)
        delta = (1 - delta**2)**2

    h = np.zeros(n)
    for x_idx, x_val in enumerate(x):
        r2 = np.array([v*v for i, v in enumerate(residuals) if x[i] == x_val])
        n = len(r2)
        se = sqrt(mean(r2)) / sqrt(len(r2))
        h[x_idx] = se * t._ppf((1+confidence)/2., n-1)

    return yest, yest-h, yest+h

def avg_lines(x, y, confidence=0.95):
    n = len(x)
    mean = np.zeros(n)
    lower = np.zeros(n)
    upper = np.zeros(n)

    for x_idx, x_val in enumerate(x):
        ys = np.array([v for i,v in enumerate(y) if x[i] == x_val])
        m,l,u = mean_confidence_interval(ys)
        mean[x_idx] = m
        lower[x_idx] = l
        upper[x_idx] = u

    return mean, lower, upper


def weighted_choice(choices):
    """
    Given a list of tuples [(val, prob),...(val, prob)] this function returns a
    randomly chosen value where the choice is weighted by prob.
    """
    total = sum(w for c, w in choices)
    r = uniform(0, total)
    upto = 0
    for c, w in choices:
       if upto + w > r:
          return c
       upto += w
    assert False, "Shouldn't get here"

def mean_confidence_interval(data, confidence=0.95):
    """
    Given a list or vector of data, this returns the mean, lower, and upper
    confidence intervals to the level of confidence specified (default = 95%
    confidence interval).
    """
    a = 1.0*np.array(data)
    n = len(a)
    m, se = np.mean(a), sem(a)
    h = se * t._ppf((1+confidence)/2., n-1)
    return m, m-h, m+h



from concept_formation.trestle import TrestleTree
from concept_formation.cluster import cluster

# Data is stored in a list of dictionaries where values can be either nominal,
# numeric, component, or relational.

data = [{'f1': 'v1', #nominal value
         'f2': 2.6, #numeric value
         'f3': {'sub-feature1': 'v1'}, # component value
         'f4': {'sub-feature1': 'v1'}, # component value
         'f5': ['some-relation', 'f3', 'f4'] #relational value
        },
        {'f1': 'v1', #nominal value
         'f2': 2.8, #numeric value
         'f3': {'sub-feature1': 'v2'}, # component value
         'f4': {'sub-feature1': 'v1'}, # component value
         'f5': ['some-relation', 'f3', 'f4'] #relational value
        }]

# Data can be clustered with a TrestleTree, which supports all data types or
# with a specific tree (CobwebTree or Cobweb3Tree) that supports subsets of
# datatypes.
tree = TrestleTree()
tree.fit(data)

# Trees can be printed in plaintext or exported in JSON format
print(tree)
print(tree.root.output_json())

# Trees can also be used to predict missing attributes of new data points.
new = {'f2': 2.6, 'f3': {'sub-feature1': 'v1'}, 'f4': {'sub-feature1': 'v1'},
       'f5': ['some-relation', 'f3', 'f4']}
concept = tree.categorize(new)
print(concept.predict('f1'))

# Trees can be used to produce flat clusterings
new_tree = TrestleTree()
clustering = cluster(new_tree, data, minsplit=1, maxsplit=1)
print(clustering)








from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division
from os.path import dirname
from os.path import join
import json

def load_json(filename):
    """
    Loads a json file and returns a python object generated from parsing the
    json.
    """
    module_path = dirname(__file__)
    with open(join(module_path, 'data_files', filename)) as dat:
        output = json.load(dat)
    return output

def load_file(filename):
    """
    Reads the rows of a file and returns them as an array.
    """
    module_path = dirname(__file__)
    with open(join(module_path, 'data_files', filename)) as dat:
        output = [row for row in dat]
    return output

def load_iris():
    """
    Load the iris dataset.
    """
    return load_json('iris.json')

def load_mushroom():
    """
    Load the mushroom dataset.
    """
    return load_json('mushrooms.json')

def load_rb_com_11():
    """
    Load the RumbleBlocks, Center of Mass Level 11, dataset.
    """
    return load_json('rb_com_11_continuous.json')

def load_rb_s_07():
    """
    Load the RumbleBlocks, Symmetry Level 7, dataset.
    """
    return load_json('rb_s_07_continuous.json')

def load_rb_s_13():
    """
    Load the RumbleBlocks, Symmetry Level 13, dataset.
    """
    return load_json('rb_s_13_continuous.json')

def load_rb_wb_03():
    """
    Load the RumbleBlocks, Wide Base Level 03, dataset.
    """
    return load_json('rb_wb_03_continuous.json')

def load_rb_s_07_human_predictions():
    """
    Load the Human Predictions Data for the RumbleBlocks, Symmetry Level 7,
    dataset.
    """
    return load_file('human_s_07_success_predictions.csv')



from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division
from collections import deque
from heapq import heappush
from heapq import heappop
from heapq import heapify
import random

class Node:
    
    def __init__(self, state, parent=None, action=None, cost=0, depth=0,
                 extra=None):
        self.state = state
        self.extra = extra
        self.parent = parent
        self.action = action
        self.cost = cost
        self.depth = depth

    def getSolution(self):
        actions = []
        current = self
        while current.parent:
            actions.append(current.action)
            current = current.parent
        actions.reverse()
        return actions

    def __str__(self):
        return str(self.state) + str(self.extra)

    def __repr__(self):
        return repr(self.state)

    def __hash__(self):
        return hash(self.state)

    def __eq__(self, other):
        return isinstance(other, Node) and self.state == other.state

def IDDFS(initial, successorFn, goalTestFn, initialDepthLimit=1):
    """
    Depth Limited Depth First Search
    """
    depthLimit = initialDepthLimit
    maxDepth = False
    nodeCount = 0
    successorFn = successorFn
    goalTestFn = goalTestFn

    while not maxDepth:
        maxDepth = True
        fringe = deque()
        getNext = fringe.pop
        fringe.append(initial)

        while fringe:
            current = getNext()
            nodeCount += 1
            if goalTestFn(current):
                print("Succeeded!")
                print("Nodes evaluated: %i" % nodeCount)
                yield current

            if current.depth < depthLimit:
                fringe.extend(successorFn(current))
            else:
                maxDepth = False
        depthLimit += 1
        print("Increasing depth limit to: %i" % depthLimit)

    print("Failed")
    print("Nodes evaluated: %i" % nodeCount)

def DLDFS(initial, successorFn, goalTestFn, depthLimit=float('inf')):
    """
    Depth Limited Depth First Search
    """
    fringe = deque()
    getNext = fringe.pop
    nodeCount = 0
    fringe.append(initial)
    successorFn = successorFn
    goalTestFn = goalTestFn

    while fringe:
        current = getNext()

        nodeCount += 1
        if goalTestFn(current):
            print("Succeeded!")
            print("Nodes evaluated: %i" % nodeCount)
            yield current

        if current.depth < depthLimit:
            fringe.extend(successorFn(current))

    print("Failed")
    print("Nodes evaluated: %i" % nodeCount)

def DepthFS(initial, successorFn, goalTestFn):
    """
    Depth first search
    """
    fringe = deque()
    getNext = fringe.pop
    nodeCount = 0
    fringe.append(initial)
    successorFn = successorFn
    goalTestFn = goalTestFn

    while fringe:
        current = getNext()

        nodeCount += 1
        if goalTestFn(current):
            print("Succeeded!")
            print("Nodes evaluated: %i" % nodeCount)
            yield current

        fringe.extend(successorFn(current))

    print("Failed")
    print("Nodes evaluated: %i" % nodeCount)

def DepthFGS(initial, successorFn, goalTestFn):
    """
    Depth first grid search (removes duplicates)
    """
    nodeCount = 0
    fringe = deque()
    closedList = set()
    openList = set()

    fringe.append(initial)
    openList.add(initial)

    while fringe:
        current = fringe.pop()
        openList.remove(current)
        closedList.add(current)

        if goalTestFn(current):
            #print("Succeeded!")
            #print("Nodes evaluated: %i" % nodeCount)
            yield current

        # Trick to push the looping into C execution
        added = [fringe.append(s) or openList.add(s) for s in
                 successorFn(current) if s not in closedList and s not in
                 openList]
        nodeCount += len(added)
        #for s in successorFn(current):
        #    nodeCount += 1
        #    if s not in closedList and s not in openList:
        #        fringe.append(s)
        #        openList.add(s)

    #print("Failed")
    #print("Nodes evaluated: %i" % nodeCount)

def BreadthFS(initial, successorFn, goalTestFn):
    """
    Breadth First search
    """
    nodeCount = 1
    fringe = deque()
    fringe.append(initial)

    while fringe:
        current = fringe.popleft()

        nodeCount += 1
        if goalTestFn(current):
            print("Succeeded!")
            print("Nodes evaluated: %i" % nodeCount)
            yield current

        for s in successorFn(current):
            nodeCount += 1
            fringe.append(s)

    print("Failed")
    print("Nodes evaluated: %i" % nodeCount)

def BreadthFGS(initial, successorFn, goalTestFn):
    """
    Breadth First Graph Search
    """
    nodeCount = 1
    fringe = deque()
    closedList = set()
    openList = set()
    fringe.append(initial)
    openList.add(initial)

    while fringe:
        current = fringe.popleft()
        openList.remove(current)
        closedList.add(current)

        if goalTestFn(current):
            #print("Succeeded!")
            #print("Nodes evaluated: %i" % nodeCount)
            yield current

        added = [fringe.append(s) or openList.add(s) for s in
                 successorFn(current) if s not in closedList and s not in
                 openList]
        nodeCount += len(added)
        #for s in successorFn(current):
        #    nodeCount += 1
        #    if s not in closedList and s not in openList:
        #        fringe.append(s)
        #        openList.add(s)

    #print("Failed")
    #print("Nodes evaluated: %i" % nodeCount)

def IDBFS(initial, successorFn, goalTestFn, heuristicFn,
          initialCostLimit=1, costInc=1):
    """
    Cost limited Best first search
    """
    nodeCount = 1
    successorFn = successorFn
    heuristicFn = heuristicFn
    goalTestFn = goalTestFn

    costMax = False
    costLimit = initialCostLimit

    while not costMax:
        costMax = True
        fringe = []
        openList = {}
        heappush(fringe, (0, nodeCount, initial))
        openList[initial] = 0

        while fringe:
            cost, counter, current = heappop(fringe)
            del openList[current]

            if goalTestFn(current):
                print("Succeeded!")
                print("Nodes evaluated: %i" % nodeCount)
                yield current

            for s in successorFn(current):
                nodeCount += 1
                sCost = s.cost + heuristicFn(s)
                if sCost <= costLimit:
                    if s in openList:
                        if sCost < openList[s]:
                            #fringe.remove((openList[s], s))
                            fringe = [e for e in fringe if e[2] != s]
                            heapify(fringe)
                            openList[s] = sCost
                            heappush(fringe, (sCost, nodeCount, s))
                    else:
                        openList[s] = sCost
                        heappush(fringe, (sCost, nodeCount, s))
                else:
                    costMax = False
        costLimit += costInc
        #print("Increasing cost limit to: %i" % costLimit)

    print("Failed")
    print("Nodes evaluated: %i" % nodeCount)

def CLBFS(initial, successorFn, goalTestFn, heuristicFn,
          costLimit=float('inf')):
    """
    Cost limited Best first search
    """
    nodeCount = 1

    fringe = []
    openList = {}
    heappush(fringe, (0, nodeCount, initial))
    openList[initial] = 0

    while fringe:
        cost, counter, current = heappop(fringe)
        del openList[current]

        if goalTestFn(current):
            print("Succeeded!")
            print("Nodes evaluated: %i" % nodeCount)
            yield current

        for s in successorFn(current):
            nodeCount += 1
            sCost = s.cost + heuristicFn(s)
            if sCost <= costLimit:
                if s in openList:
                    if sCost < openList[s]:
                        #fringe.remove((openList[s], len(fringe), s))
                        fringe = [e for e in fringe if e[2] != s]
        
                        heapify(fringe)
                        openList[s] = sCost
                        heappush(fringe, (sCost, nodeCount, s))
                else:
                    openList[s] = sCost
                    heappush(fringe, (sCost, nodeCount, s))
            else:
                print("Cost limit hit")

    print("Failed")
    print("Nodes evaluated: %i" % nodeCount)

def BestFS(initial, successorFn, goalTestFn, heuristicFn):
    """
    Best first search
    """
    nodeCount = 1
    fringe = []
    openList = {}
    heappush(fringe, (0, nodeCount, initial))
    openList[initial] = 0.0

    while fringe:
        cost, counter, current = heappop(fringe)
        del openList[current]

        if goalTestFn(current):
            print("Succeeded!")
            print("Nodes evaluated: %i" % nodeCount)
            yield current

        for s in successorFn(current):
            nodeCount += 1
            sCost = s.cost + heuristicFn(s)
            if s in openList:
                if sCost < openList[s]:
                    #fringe.remove((openList[s], ), s))
                    fringe = [e for e in fringe if e[2] != s]
                    heapify(fringe)
                    openList[s] = sCost
                    heappush(fringe, (sCost, nodeCount, s))
            else:
                openList[s] = sCost
                heappush(fringe, (sCost, nodeCount, s))

    print("Failed")
    print("Nodes evaluated: %i" % nodeCount)

def BestFGS(initial, successorFn, goalTestFn, heuristicFn):
    nodeCount = 1
    fringe = []
    closedList = {}
    openList = {}
    heappush(fringe, (0, nodeCount, initial))
    openList[initial] = 0.0

    while fringe:
        cost, counter, current = heappop(fringe)
        del openList[current]
        closedList[current] = cost

        if goalTestFn(current):
            #print("Succeeded!")
            #print("Nodes evaluated: %i" % nodeCount)
            yield current

        for s in successorFn(current):
            nodeCount += 1
            sCost = s.cost + heuristicFn(s)
            if s in openList:
                if sCost < openList[s]:
                    fringe = [e for e in fringe if e[2] != s]
                    heapify(fringe)
                    openList[s] = sCost
                    heappush(fringe, (sCost, nodeCount, s))
            elif s in closedList:
                if sCost < closedList[s]:
                    del closedList[s]
                    openList[s] = sCost
                    heappush(fringe, (sCost, nodeCount, s))
            else:
                openList[s] = sCost
                heappush(fringe, (sCost, nodeCount, s))

    #print("Failed")
    #print("Nodes evaluated: %i" % nodeCount)

def BeamS(initial, successorFn, goalTestFn, heuristicFn, initialBeamWidth=1):
    """
    Beam Search (allows duplicates)
    """
    nodeCount = 1
    beamMax = False
    beamWidth = initialBeamWidth
    while not beamMax:
        beamMax = True
        fringe = []
        openList = {}
        heappush(fringe, (0, nodeCount, initial))
        openList[initial] = 0

        while fringe:
            cost, counter, current = heappop(fringe)

            if goalTestFn(current):
                #print("Succeeded!")
                #print("Nodes evaluated: %i" % nodeCount)
                yield current

            for s in successorFn(current):
                nodeCount += 1
                sCost = s.cost + heuristicFn(s)
                if s in openList:
                    if sCost < openList[s]:
                        #fringe.remove((openList[s], s))
                        fringe = [e for e in fringe if e[2] != s]
                        heapify(fringe)
                        openList[s] = sCost
                        heappush(fringe, (sCost, nodeCount, s))
                else:
                    openList[s] = sCost
                    heappush(fringe, (sCost, nodeCount, s))

            if len(fringe) > beamWidth:
                best = []
                openList = {}
                for i in range(beamWidth):
                    if fringe:
                        c, s = heappop(fringe)
                        openList[s] = c
                        heappush(best, (cost, nodeCount, s))
                fringe = best
                beamMax = False
        
        beamWidth += 1
        #print("Increasing beam width to: %i" % beamWidth)

    #print("Failed")
    #print("Nodes evaluated: %i" % nodeCount)

def BeamGS(initial, successorFn, goalTestFn, heuristicFn, initialBeamWidth=1):
    """
    Beam Grid Search (no duplicates)
    """
    nodeCount = 1

    beamMax = False
    beamWidth = initialBeamWidth
    while not beamMax:
        beamMax = True
        fringe = []
        openList = {}
        closedList = {}
        heappush(fringe, (0, nodeCount, initial))
        openList[initial] = 0

        while fringe:
            cost, counter, current = heappop(fringe)
            del openList[current]
            closedList[current] = cost

            if goalTestFn(current):
                #print("Succeeded!")
                #print("Nodes evaluated: %i" % nodeCount)
                yield current

            for s in successorFn(current):
                nodeCount += 1
                sCost = s.cost + heuristicFn(s)
                if s in openList:
                    if sCost < openList[s]:
                        #fringe.remove((openList[s], s))
                        fringe = [e for e in fringe if e[2] != s]
                        heapify(fringe)
                        openList[s] = sCost
                        heappush(fringe, (sCost, nodeCount, s))
                elif s in closedList:
                    if sCost < closedList[s]:
                        del closedList[s]
                        openList[s] = sCost
                        heappush(fringe, (sCost, nodeCount, s))
                else:
                    openList[s] = sCost
                    heappush(fringe, (sCost, nodeCount, s))

            if len(fringe) > beamWidth:
                best = []
                openList = {}
                for i in range(beamWidth):
                    if fringe:
                        c, count, s = heappop(fringe)
                        openList[s] = c
                        heappush(best, (c, count, s))
                fringe = best
                beamMax = False
        
        beamWidth += 1
        print("Increasing beam width to: %i" % beamWidth)

    print("Failed")
    print("Nodes evaluated: %i" % nodeCount)

class eightPuzzle:

    def __init__(self):
        self.state = (0,1,2,3,4,5,6,7,8)

    def __hash__(self):
        return hash(self.state)

    def __eq__(self, other):
        if isinstance(other, eightPuzzle):
            return self.state == other.state
        return False

    def copy(self):
        new = eightPuzzle()
        new.state = tuple([i for i in self.state])
        return new

    def randomize(self):
        state = [0,1,2,3,4,5,6,7,8]
        self.state = tuple(state)

        for i in range(100):
            self.executeAction(random.choice([a for a in self.legalActions()]))
 
    def __repr__(self):
        return str(self)

    def __str__(self):
        return "%i%i%i\n%i%i%i\n%i%i%i" % (self.state[0], self.state[1],
                                           self.state[2], self.state[3],
                                           self.state[4], self.state[5],
                                           self.state[6], self.state[7],
                                           self.state[8])

    def executeAction(self, action):
        zeroIndex = self.state.index(0)
        successor = list(self.state)

        if action == 'up':
            successor[zeroIndex] = successor[zeroIndex + 3]
            successor[zeroIndex + 3] = 0
        elif action == 'left':
            successor[zeroIndex] = successor[zeroIndex + 1]
            successor[zeroIndex + 1] = 0
        elif action == 'right':
            successor[zeroIndex] = successor[zeroIndex - 1]
            successor[zeroIndex - 1] = 0
        elif action == 'down':
            successor[zeroIndex] = successor[zeroIndex - 3]
            successor[zeroIndex - 3] = 0

        self.state = tuple(successor)

    def legalActions(self):
        zeroIndex = self.state.index(0)

        if zeroIndex in set([0,1,2,3,4,5]):
            yield "up"
        if zeroIndex in set([0,3,6,1,4,7]):
            yield "left"
        if zeroIndex in set([2,5,8,1,4,7]):
            yield "right"
        if zeroIndex in set([3,4,5,6,7,8]):
            yield "down"

def successorFn8Puzzle(node):
    for action in node.state.legalActions():
        newState = node.state.copy()
        newState.executeAction(action)
        yield Node(newState, node, action, node.cost+1, node.depth+1)

def heuristicFn8Puzzle(node):
    """
    Currently the misplaced tile count
    """
    goal = eightPuzzle()
    h = 0
    for i,v in enumerate(node.state.state):
        if node.state.state[i] != goal.state[i]:
            h += 1
    return h

def goalTestFn8Puzzle(node):
    goal = eightPuzzle()
    return node.state == goal

if __name__ == "__main__":

    puzzle = eightPuzzle()
    puzzle.randomize()
    #puzzle.executeAction('left')
    #puzzle.executeAction('up')
    #puzzle.executeAction('up')

    initial = puzzle
    print("INITIAL:")
    print(puzzle)
    print()

    print("BREADTH FIRST GRAPH SEARCH")
    sol = next(BreadthFGS(Node(initial), successorFn8Puzzle, goalTestFn8Puzzle))
    print("Solution Length = %i" % len(sol.getSolution()))
    print()

    print("DEPTH FIRST GRAPH SEARCH")
    sol = next(DepthFGS(Node(initial), successorFn8Puzzle, goalTestFn8Puzzle))
    print("Solution Length = %i" % len(sol.getSolution()))
    print()
    
    print("BEST FIRST GRAPH SEARCH")
    sol = next(BestFGS(Node(initial), successorFn8Puzzle, goalTestFn8Puzzle,
                  heuristicFn8Puzzle))
    print("Solution Length = %i" % len(sol.getSolution()))
    print()

    print("BEAM GRAPH SEARCH")
    sol = next(BeamGS(Node(initial), successorFn8Puzzle, goalTestFn8Puzzle,
                  heuristicFn8Puzzle, 10))
    print("Solution Length = %i" % len(sol.getSolution()))
    print()

#    print("ITERATIVE DEEPENING DEPTH FIRST SEARCH")
#    sol = next(IDDFS(Node(initial), successorFn8Puzzle, goalTestFn8Puzzle))
#    print("Solution Length = %i" % len(sol.getSolution()))
#    print()
#
    print("ITERATIVE DEEPENING BEST FIRST SEARCH")
    sol = next(IDBFS(Node(initial), successorFn8Puzzle, goalTestFn8Puzzle,
                  heuristicFn8Puzzle))
    print("Solution Length = %i" % len(sol.getSolution()))
    print()

    #print("BEAM SEARCH")
    #sol = next(BeamS(Node(initial), successorFn8Puzzle, goalTestFn8Puzzle,
    #              heuristicFn8Puzzle))
    #print("Solution Length = %i" % len(sol.getSolution()))
    #print()

    #print("BREADTH FIRST SEARCH")
    #sol = next(BreadthFS(Node(initial), successorFn8Puzzle, goalTestFn8Puzzle))
    #print("Solution Length = %i" % len(sol.getSolution()))
    #print()

    #print("DEPTH FIRST SEARCH")
    #sol = next(DepthFS(Node(initial), successorFn8Puzzle, goalTestFn8Puzzle))
    #print("Solution Length = %i" % len(sol.getSolution()))
    #print()






from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division
from random import normalvariate
from numbers import Number
from math import sqrt
from math import pi
from math import exp

from concept_formation.utils import c4
from concept_formation.cobweb import CobwebNode
from concept_formation.cobweb import CobwebTree

# Magnify the continuous values so that the just noticable difference
# of the noramlized values is equal to acuity (i.e., a JND of 8% means that the
# system is able to perceive a difference of 0.08 after normalizing the data to
# have std 1). Smaller JND values tend to make categorization slower because
# the tree will have greater depth due to being able to more finely
# discriminate between values.
just_noticable_difference = 0.08
scale_num_std = 1 /(just_noticable_difference * sqrt(2 * pi))
scale_proportion = 1 / scale_num_std

class Cobweb3Tree(CobwebTree):

    def __init__(self, alpha=0.001, scaling=True):
        """
        The tree constructor. 

        The alpha parameter is the parameter used for laplacian smoothing. The
        higher the value, the higher the prior that all attributes/values are
        equally likely. By default a minor smoothing is used: 0.001.

        The scaling parameter determines whether online normalization of
        continuous attributes is used. By default scaling is used. Scaling
        divides the std of each attribute by the std of the attribute in the
        root node. 
        """
        self.root = Cobweb3Node()
        self.root.root = self.root
        self.root.alpha = alpha
        self.root.scaling = scaling

class Cobweb3Node(CobwebNode):

    # Smallest possible acuity. Below this probabilities will exceed 1.0
    acuity = 1.0 / sqrt(2.0 * pi)

    def increment_counts(self, instance):
        """
        A modified version of increment counts that handles floats properly
        """
        self.count += 1 
            
        for attr in instance:
            if isinstance(instance[attr], Number):
                if (attr not in self.av_counts or 
                    not isinstance(self.av_counts[attr], ContinuousValue)):
                    # TODO currently overrides nominals if a float comes in.
                    self.av_counts[attr] = ContinuousValue()
                self.av_counts[attr].update(instance[attr])
            else:
                self.av_counts[attr] = self.av_counts.setdefault(attr,{})
                self.av_counts[attr][instance[attr]] = (self.av_counts[attr].get(instance[attr], 0) + 1)

    def update_counts_from_node(self, node):
        """
        modified to handle floats
        Increments the counts of the current node by the amount in the specified
        node.
        """
        self.count += node.count
        for attr in node.av_counts:
            if isinstance(node.av_counts[attr], ContinuousValue):
                if (attr not in self.av_counts or 
                    not isinstance(self.av_counts[attr], ContinuousValue)):
                    # TODO currently overrides nominals if a float comes in.
                    self.av_counts[attr] = ContinuousValue()
                self.av_counts[attr].combine(node.av_counts[attr])
            else:
                for val in node.av_counts[attr]:
                    self.av_counts[attr] = self.av_counts.setdefault(attr,{})
                    self.av_counts[attr][val] = (self.av_counts[attr].get(val,0) +
                                         node.av_counts[attr][val])
    
    def attr_val_guess_gain(self, attr, val):
        """
        Returns the gain in number of correct guesses if a particular attr/val
        was added to a concept.
        """
        if attr[0] == "_":
            return 0.0
        elif attr not in self.av_counts:
            return 0.0
        elif isinstance(self.av_counts[attr], ContinuousValue):
            if self.root.scaling:
                scale = scale_proportion * self.root.av_counts[attr].unbiased_std()
            else:
                scale = 1.0

            # TODO consider incorporating laplace smoothing (alpha).
            before_std = max(self.av_counts[attr].scaled_unbiased_std(scale), self.acuity)
            before_prob = ((1.0 * self.av_counts[attr].num) / (self.count + 1.0))
            before_count = ((before_prob * before_prob) * 
                            (1.0 / (2.0 * sqrt(pi) * before_std)))

            temp = self.av_counts[attr].copy()
            temp.update(val)
            after_std = max(temp.scaled_unbiased_std(scale), self.acuity)
            after_prob = ((1.0 + self.av_counts[attr].num) / (self.count + 1.0))
            after_count = ((after_prob * after_prob) * 
                            (1.0 / (2.0 * sqrt(pi) * after_std)))
            return after_count - before_count
        elif val not in self.av_counts[attr]:
            return 0.0
        else:
            before_prob = (self.av_counts[attr][val] / (self.count + 1.0))
            after_prob = (self.av_counts[attr][val] + 1) / (self.count + 1.0)

            return (after_prob * after_prob) - (before_prob * before_prob)

    def expected_correct_guesses(self):
        """
        Computes the number of attribute values that would be correctly guessed
        in the current concept. This extension supports both nominal and
        numeric attribute values. 
        
        The typical cobweb 3 calculation for correct guesses is:

        .. math::

            P(A_i = V_{ij})^2 = \\frac{1}{2 * \\sqrt{\\pi} * \\sigma}

        However, this does not take into account situations when 
        :math:`P(A_i) \\neq 1.0`.

        To account for this we use a modified equation:

        .. math::

            P(A_i = V_{ij})^2 = P(A_i)^2 * \\frac{1}{2 * \\sqrt{\\pi} * \\sigma}

        """
        correct_guesses = 0.0

        for attr in self.root.av_counts:
            if attr[0] == "_":
                continue
            elif isinstance(self.root.av_counts[attr], ContinuousValue):
                n_values = 2
                if attr not in self.av_counts :
                    prob = 0
                    if self.root.alpha > 0:
                        prob = self.root.alpha / (self.root.alpha * n_values)
                    val_count = 0
                else:
                    val_count = self.av_counts[attr].num

                    if self.root.scaling:
                        scale = scale_proportion * self.root.av_counts[attr].unbiased_std()
                    else:
                        scale = 1.0

                    std = max(self.av_counts[attr].scaled_unbiased_std(scale),
                              self.acuity)
                    prob_attr = ((1.0 * self.av_counts[attr].num + self.root.alpha) /
                                 (self.count + self.root.alpha * n_values ))
                    correct_guesses += ((prob_attr * prob_attr) * 
                                        (1.0 / (2.0 * sqrt(pi) * std)))

                #Factors in the probability mass of missing values
                prob = ((self.count - val_count + self.root.alpha) / (1.0 * self.count +
                                                            self.root.alpha * n_values))
                correct_guesses += (prob * prob)

            else:
                val_count = 0
                n_values = len(self.root.av_counts[attr]) + 1
                for val in self.root.av_counts[attr]:
                    if attr not in self.av_counts or val not in self.av_counts[attr]:
                        prob = 0
                        if self.root.alpha > 0:
                            prob = self.root.alpha / (self.root.alpha * n_values)
                    else:
                        val_count += self.av_counts[attr][val]
                        prob = ((self.av_counts[attr][val] + self.root.alpha) / (1.0 * self.count + 
                                                                       self.root.alpha * n_values))
                    correct_guesses += (prob * prob)

                #Factors in the probability mass of missing values
                prob = ((self.count - val_count + self.root.alpha) /
                        (1.0*self.count + self.root.alpha *
                                                    n_values))
                correct_guesses += (prob * prob)

        return correct_guesses

    def pretty_print(self, depth=0):
        """
        Prints the categorization tree.
        """
        ret = str(('\t' * depth) + "|-")

        attributes = []

        for attr in self.av_counts:
            if isinstance(self.av_counts[attr], ContinuousValue):
                attributes.append("'%s': { %0.3f (%0.3f) [%i] }" % (attr,
                                                                    self.av_counts[attr].mean,
                                                                    max(self.acuity,
                                                                        self.av_counts[attr].unbiased_std()),
                                                                    self.av_counts[attr].num))
            else:
                values = []

                for val in self.av_counts[attr]:
                    values.append("'" + str(val) + "': " +
                                  str(self.av_counts[attr][val]))

                attributes.append("'" + attr + "': {" + ", ".join(values) + "}")
                  
        ret += "{" + ", ".join(attributes) + "}: " + str(self.count) + '\n'
        
        for c in self.children:
            ret += c.pretty_print(depth+1)

        return ret

    def sample(self, attr):
        """
        Uses the probability table to sample a value for an attribute. This can
        be useful when wanting to complete missing attributes of objects.
        """
        if attr not in self.root.av_counts:
            return None

        if isinstance(self.root.av_counts[attr], ContinuousValue):
            n_values = 2
            prob_attr = ((1.0 * self.av_counts[attr].num + self.root.alpha) /
                         (self.count + self.root.alpha * n_values ))

            if prob_attr < 0.5:
                return None

            return normalvariate(self.av_counts[attr].mean,
                                 self.av_counts[attr].unbiased_std())
        else:
            return super(Cobweb3Node, self).sample(attr)

    def predict(self, attr):
        """
        Predicts the value of an attribute, in the case of nominal attributes
        the most likely value is chosen. In the case of continuous attributes,
        the mean value is chosen. 
        """
        if attr not in self.root.av_counts:
            return None

        if isinstance(self.root.av_counts[attr], ContinuousValue):
            n_values = 2
            prob_attr = ((1.0 * self.av_counts[attr].num + self.root.alpha) /
                         (self.count + self.root.alpha * n_values ))

            if prob_attr < 0.5:
                return None

            return self.av_counts[attr].mean
        else:
            return super(Cobweb3Node, self).predict(attr)

    def get_probability(self, attr, val):
        """
        Gets the probability of a particular attribute value. This takes into
        account the possibility that a value is missing.
        """
        if attr not in self.root.av_counts:
            return 0.0

        if isinstance(self.root.av_counts[attr], ContinuousValue):
            n_values = 2
            prob_attr = ((1.0 * self.av_counts[attr].num + self.root.alpha) /
                         (self.count + self.root.alpha * n_values ))

            if val is None:
                return 1 - prob_attr

            if self.root.scaling:
                scale = scale_proportion * self.root.av_counts[attr].unbiased_std()
                if scale == 0:
                    scale = 1
                shift = self.root.av_counts[attr].mean
                val = (val - shift) / scale
            else:
                scale = 1.0
                shift = 0.0

            mean = (self.av_counts[attr].mean - shift) / scale
            std = max(self.av_counts[attr].scaled_unbiased_std(scale),
                      self.acuity)

            return (prob_attr * 
                    (1.0 / (std * sqrt(2 * pi))) * 
                    exp(-((val - mean) * (val - mean)) / (2.0 * std * std)))

        else:
            return super(Cobweb3Node, self).get_probability(attr, val)

    def output_json(self):
        """
        A modification of the cobweb output json to handle numeric values.
        """
        output = {}
        if "_guid" in self.av_counts:
            for guid in self.av_counts['_guid']:
                output['guid'] = guid
        output["name"] = "Concept" + self.concept_id
        output["size"] = self.count
        output["children"] = []

        temp = {}
        for attr in self.av_counts:
            #float_vals = []
            if isinstance(self.av_counts[attr], ContinuousValue):
                temp[str(attr) + " = " + str(self.av_counts[attr])] = self.av_counts[attr].num
            else:
                for value in self.av_counts[attr]:
                    temp[str(attr) + " = " + str(value)] = self.av_counts[attr][value]

        for child in self.children:
            output["children"].append(child.output_json())

        output["counts"] = temp

        return output

class ContinuousValue():
    """ 
    This class scores the number of samples, the mean of the samples, and the
    squared error of the samples. It can be used to perform incremental
    estimation of the mean, std, and unbiased std.
    """

    def __init__(self):
        """
        Initializes the number of values, the mean of the values, and the
        squared errors of the values to 0.
        """
        self.num = 0.0
        self.mean = 0.0
        self.meanSq = 0.0

    def __len__(self):
        return 1

    def copy(self):
        """
        Returns a deep copy of itself.
        """
        v = ContinuousValue()
        v.num = self.num
        v.mean = self.mean
        v.meanSq = self.meanSq
        return v

    def unbiased_mean(self):
        """
        Returns the mean value.
        """
        return self.mean

    def scaled_unbiased_mean(self, shift, scale):
        """
        Returns the mean shifted and scaled.
        """
        if scale <= 0:
            scale = 1
        return (self.mean - shift) / scale

    def biased_std(self):
        """
        Returns a biased estimate of the std (i.e., the sample std)
        """
        return sqrt(self.meanSq / (self.num))

    def unbiased_std(self):
        """
        Returns an unbiased estimate of the std that uses Bessel's correction
        and Cochran's theorem: 
        `<https://en.wikipedia.org/wiki/Unbiased_estimation_of_standard_deviation#Bias_correction>`_
        """
        if self.num < 2:
            return 0.0
        return sqrt(self.meanSq / (self.num - 1)) / c4(self.num)

    def scaled_unbiased_std(self, scale):
        """
        Returns an unbiased estimate of the std (see comments on unbiased_std),
        but also adjusts the std given a scale parameter. This is used to
        return std values that have been normalized by some value.

        For edge cases, if scale is less than or equal to 0, then scaling is
        disabled (i.e., scale = 1.0).
        """
        if scale <= 0:
            scale = 1.0
        return self.unbiased_std() / scale

    def __hash__(self):
        """
        This hashing function returns the hash of a constant string, so that
        all lookups of a continuous value in a dictionary get mapped to the
        same entry. 
        """
        return hash("#ContinuousValue#")

    def __repr__(self):
        """
        The representation of a continuous value.
        """
        return repr(self.num) + repr(self.mean) + repr(self.meanSq)

    def __str__(self):
        """
        The string format for a continuous value."
        """
        return "%0.4f (%0.4f) [%i]" % (self.mean, self.unbiased_std(), self.num)

    def update_batch(self, data):
        """
        Calls the update function on every value in the given dataset
        """
        for x in data:
            self.update(x)

    def update(self, x):
        """
        Incrementally update the mean and squared mean error (meanSq) values in
        an efficient and practical (no precision problems) way. This uses and
        algorithm by Knuth, which I found here:
        `<https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance>`_
        """
        self.num += 1
        delta = x - self.mean 
        self.mean += delta / self.num
        self.meanSq += delta * (x - self.mean)

    def combine(self, other):
        """
        Combine two clusters of means and squared mean error (meanSq) values in
        an efficient and practical (no precision problems) way. This uses the
        parallel algorithm by Chan et al. found at:
        `<https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm>`_
            """
        if not isinstance(other, ContinuousValue):
            raise ValueError("Can only merge 2 continuous values.")
        delta = other.mean - self.mean
        self.meanSq = (self.meanSq + other.meanSq + delta * delta * 
                       ((self.num * other.num) / (self.num + other.num)))
        self.mean = ((self.num * self.mean + other.num * other.mean) / 
                     (self.num + other.num))
        self.num += other.num



from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division
import json
import re
import copy

def cluster_iter(tree, instances, minsplit=1, maxsplit=100000,  mod=True):
    """Categorize the list of instances into the tree and return an iterator that
    can  be used to iteratively return clusterings based on splitting nodes of
    the tree.

    The first clustering is derived by splitting the root, then each subsequent
    clustering is based on splitting the least coupled cluster (in terms of
    category utility). This process may halt early if it reaches the case where
    there are no more clusters to split (each instance in its own cluster).
    Because splitting is a modifying operation, a deepcopy of the tree is made
    before creating the iterator.

    :param tree: A category tree to generate clusters from, it can be pre-trained or newly created.
    :param instances: A list of instances
    :param minsplit: The minimum number of splits to perform on the tree (default 1)
    :param maxsplit: the maximum number of splits to perform on the tree (default 100000)
    :param mod: A flag determine if instances will be fit (i.e. modifying knoweldge) or categorized (i.e. not modifiying knowledge)
    :type tree: CobwebTree, Cobweb3Tree, or TrestleTree
    :type instances: list
    :type minsplit: int
    :type maxsplit: int
    :type mod: bool
    :returns: an iterator of clusterings based on a number of splits between minsplit and maxsplit
    :rtype: iterator

    .. warning:: minsplit must be >=1 and maxsplit must be >= minsplit
    """
    if minsplit < 1: 
        raise ValueError("minsplit must be >= 1") 
    if minsplit > maxsplit: 
        raise ValueError("maxsplit must be >= minsplit")

    tree = copy.deepcopy(tree)

    if mod:
        temp_clusters = [tree.ifit(instance) for instance in instances]
    else:
        temp_clusters = [tree.categorize(instance) for instance in instances]
    
    for nth_split in range(1,maxsplit+1):

        if nth_split >= minsplit:
            clusters = []
            for i,c in enumerate(temp_clusters):
                while (c.parent and c.parent.parent):
                    c = c.parent
                clusters.append("Concept" + c.concept_id)
            yield clusters

        split_cus = sorted([(tree.root.cu_for_split(c) -
                             tree.root.category_utility(), i, c) for i,c in
                            enumerate(tree.root.children) if c.children])

        # Exit early, we don't need to re-reun the following part for the
        # last time through
        if not split_cus:
            break

        # Split the least cohesive cluster
        tree.root.split(split_cus[-1][2])

        nth_split+=1

def cluster(tree, instances, minsplit=1, maxsplit=1, mod=True):
    """
    Categorize the list of instances into the tree and return a list of lists of
    flat cluster labelings based on different numbers of splits.

    Keyword arguments:
    tree -- a CobwebTree, Cobweb3Tree, or TrestleTree
    instances -- a list of instance objects
    minsplit -- The minimum number of splits to perform on the tree, must be >=1
    (default 1)
    maxsplit -- the maximum number of splits to perform on the tree, must be >=1
    (default 1)
    mod -- If True instances will be fit (i.e. modyfiying knoweldge) else they
    will be categorized (i.e. not modiyfing knowledge) (default True)
    """
    return [c for c in cluster_iter(tree,instances,minsplit,maxsplit,mod)]

def k_cluster(tree,instances,k=3,mod=True):
    """
    Categorize the list of instances into the tree and return a flat cluster
    where n_clusters <= k. If a split would result in n_clusters > k then fewer
    clusters will be returned.

    Keyword arguments:
    tree -- a CobwebTree, Cobweb3Tree, or TrestleTree
    instances -- a list of instance objects
    k -- a desired number of clusters (default 3)
    mod -- If True instances will be fit (i.e. modyfiying knoweldge) else they
    will be categorized (i.e. not modiyfing knowledge) (default True)
    """

    if k < 2:
        raise ValueError("k must be >=2, all nodes in Cobweb are guaranteed to have at least 2 children.")

    clustering = ["Concept" + tree.root.concept_id for i in instances]
    for c in cluster_iter(tree, instances,mod=mod):
        if len(set(c)) > k:
            break
        clustering = c

    return clustering

def generate_d3_visualization(tree, fileName):
    """
    Export a .js file that is used to visualize the tree with d3.
    """
    fname = 'visualize/'+fileName+'.js'
    with open(fname, 'w') as f:
        f.write("var output = '"+re.sub("'", '',
                                        json.dumps(tree.root.output_json()))+"';")

def depth_labels(tree,instances,mod=True):
    """
    Categorize the list of instances into the tree and return a matrix of
    labeling of each instance based on different depth cuts of the tree.

    The returned matrix is max(conceptDepth) X len(instances). Labelings are
    ordered general to specific with final_labels[0] being the root and
    final_labels[-1] being the leaves.

    Keyword attributes:
    tree -- a CobwebTree, Cobweb3Tree, or TrestleTree
    instances -- a list of instance objects
    mod -- If True instances will be fit (i.e. modyfiying knoweldge) else they
    will be categorized (i.e. not modiyfing knowledge) (default True)
    """
    if mod:
        temp_labels = [tree.ifit(instance) for instance in instances]
    else:
        temp_labels = [tree.categorize(instance) for instance in instances]

    instance_labels = []
    max_depth = 0
    for t in temp_labels:
        labs = []
        depth = 0
        label = t
        while label.parent:
            labs.append("Concept" + label.concept_id)
            depth += 1
            label = label.parent
        labs.append("Concept" + label.concept_id)
        depth += 1
        instance_labels.append(labs)
        if depth > max_depth:
            max_depth = depth

    for f in instance_labels:
        f.reverse()
        last_label = f[-1]
        while len(f) < max_depth:
            f.append(last_label)

    final_labels = []
    for d in range(len(instance_labels[0])):
        depth_n = []
        for i in instance_labels:
            depth_n.append(i[d])
        final_labels.append(depth_n)

    return final_labels

#from trestle import TrestleTree
#import pprint
#
#with open("data_files/rb_s_07_continuous.json") as dat:
#    instances = json.load(dat)[:15]
#
#    tree = TrestleTree()
#
#    clus = cluster(tree,instances,maxsplit=5)
#    pprint.pprint(clus)
